"""Sphinx extension that defines new auto documenters with autosummary.

The :class:`AutoSummModuleDocumenter` and :class:`AutoSummClassDocumenter`
classes defined here enable an autosummary-style listing of the corresponding
members.

This extension gives also the possibility to choose which data shall be shown
and to include the docstring of the ``'__call__'`` attribute.

**Disclaimer**

Copyright 2016-2019, Philipp S. Sommer

Copyright 2020-2021, Helmholtz-Zentrum Hereon

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = "Philipp S. Sommer"
__copyright__ = (
    "2016 - 2019, Philipp S. Sommer\n"
    "2020 - 2021, Helmholtz-Zentrum Hereon"
)
__credits__ = ["Philipp S. Sommer"]
__license__ = "Apache-2.0"

__maintainer__ = "Philipp S. Sommer"
__email__ = "philipp.sommer@hereon.de"

__status__ = "Production"

from itertools import chain

from sphinx.util import logging
import re

from docutils import nodes

import sphinx

from sphinx.util.docutils import SphinxDirective

from sphinx.ext.autodoc import (
    ClassDocumenter, ModuleDocumenter, ALL, PycodeError,
    ModuleAnalyzer, AttributeDocumenter, DataDocumenter, Options,
    prepare_docstring)
import sphinx.ext.autodoc as ad

signature = Signature = None

from sphinx.ext.autodoc.directive import (
    AutodocDirective, AUTODOC_DEFAULT_OPTIONS, process_documenter_options,
    DocumenterBridge
)

try:
    from sphinx.util.inspect import signature, stringify_signature
except ImportError:
    from sphinx.ext.autodoc import Signature

sphinx_version = list(map(float, re.findall(r'\d+', sphinx.__version__)[:3]))

try:
    from sphinx.util import force_decode
except ImportError:
    # force_decode has been removed with sphinx 4.0
    def force_decode(string: str, encoding: str) -> str:
        return string


from . import _version
__version__ = _version.get_versions()['version']


logger = logging.getLogger(__name__)

#: Options of the :class:`sphinx.ext.autodoc.ModuleDocumenter` that have an
#: effect on the selection of members for the documentation
member_options = {
    'members', 'undoc-members', 'inherited-members', 'exclude-members',
    'private-members', 'special-members', 'imported-members',
    'ignore-module-all'}


if signature is not None:  # sphinx >= 2.4.0
    def process_signature(obj):
        sig = signature(obj)
        return stringify_signature(sig)
elif Signature is not None:  # sphinx >= 1.7
    def process_signature(obj):
        try:
            args = Signature(obj).format_args()
        except TypeError:
            return None
        else:
            args = args.replace('\\', '\\\\')
            return args


def list_option(option):
    """Transform a string to a list by splitting at ;;."""
    return [s.strip() for s in option.split(";;")]


def bool_option(*args):
    return "True"


def _get_arg(param, pos, default, *args, **kwargs):
    if param in kwargs:
        return kwargs[param]
    elif len(args) > pos:
        return args[pos]
    else:
        return default


class AutosummaryDocumenter(object):
    """Abstract class for for extending Documenter methods

    This classed is used as a base class for Documenters in order to provide
    the necessary methods for generating the autosummary."""

    #: List of functions that may filter the members
    filter_funcs = []

    #: Grouper functions
    grouper_funcs = []

    def __init__(self):
        raise NotImplementedError

    def get_grouped_documenters(self, all_members=False):
        """Method to return the member documenters

        This method is somewhat like a combination of the
        :meth:`sphinx.ext.autodoc.ModuleDocumenter.generate` method and the
        :meth:`sphinx.ext.autodoc.ModuleDocumenter.document_members` method.
        Hence it initializes this instance by importing the object, etc. and
        it finds the documenters to use for the autosummary option in the same
        style as the document_members does it.

        Returns
        -------
        dict
            dictionary whose keys are determined by the :attr:`member_sections`
            dictionary and whose values are lists of tuples. Each tuple
            consists of a documenter and a boolean to identify whether a module
            check should be made describes an attribute or not.

        Notes
        -----
        If a :class:`sphinx.ext.autodoc.Documenter.member_order` value is not
        in the :attr:`member_sections` dictionary, it will be put into an
        additional `Miscellaneous` section."""
        use_sections = self.options.autosummary_sections

        self.parse_name()
        self.import_object()
        # If there is no real module defined, figure out which to use.
        # The real module is used in the module analyzer to look up the module
        # where the attribute documentation would actually be found in.
        # This is used for situations where you have a module that collects the
        # functions and classes of internal submodules.
        self.real_modname = None or self.get_real_modname()

        # try to also get a source code analyzer for attribute docs
        if sphinx_version < [4, 0]:
            record_dependencies = self.directive.filename_set
        else:
            record_dependencies = self.directive.record_dependencies
        try:
            self.analyzer = ModuleAnalyzer.for_module(self.real_modname)
            # parse right now, to get PycodeErrors on parsing (results will
            # be cached anyway)
            self.analyzer.find_attr_docs()
        except PycodeError as err:
            logger.debug('[autodocsumm] module analyzer failed: %s', err)
            # no source file -- e.g. for builtin and C modules
            self.analyzer = None
            # at least add the module.__file__ as a dependency
            if hasattr(self.module, '__file__') and self.module.__file__:
                record_dependencies.add(self.module.__file__)
        else:
            record_dependencies.add(self.analyzer.srcname)

        self.env.temp_data['autodoc:module'] = self.modname
        if self.objpath:
            self.env.temp_data['autodoc:class'] = self.objpath[0]

        if not self.options.autosummary_force_inline:
            docstring = self.get_doc() or []
            autodocsumm_directive = '.. auto%ssumm::' % self.objtype
            for s in chain.from_iterable(docstring):
                if autodocsumm_directive in s:
                    return {}

        # set the members from the autosummary member options
        options_save = {}
        for option in member_options.intersection(self.option_spec):
            autopt = 'autosummary-' + option
            if getattr(self.options, autopt):
                options_save[option] = getattr(self.options, option)
                self.options[option] = getattr(self.options, autopt)

        want_all = all_members or self.options.inherited_members or \
            self.options.members is ALL
        # find out which members are documentable
        members_check_module, members = self.get_object_members(want_all)

        # remove members given by exclude-members
        if self.options.exclude_members:
            members = [(membername, member) for (membername, member) in members
                       if membername not in self.options.exclude_members]

        # document non-skipped members
        memberdocumenters = []
        registry = self.env.app.registry.documenters

        for (mname, member, isattr) in self.filter_members(members, want_all):
            classes = [cls for cls in registry.values()
                       if cls.can_document_member(member, mname, isattr, self)]
            if not classes:
                # don't know how to document this member
                continue
            # prefer the documenter with the highest priority
            classes.sort(key=lambda cls: cls.priority)
            # give explicitly separated module name, so that members
            # of inner classes can be documented
            full_mname = self.modname + '::' + \
                '.'.join(self.objpath + [mname])

            documenter = classes[-1](self.directive, full_mname, self.indent)
            memberdocumenters.append((documenter,
                                      members_check_module and not isattr))

        member_order = (
            self.options.member_order or self.env.config.autodoc_member_order
        )
        try:
            memberdocumenters = self.sort_members(
                memberdocumenters, member_order
            )
        except AttributeError:  # sphinx<3.0
            pass

        documenters = {}
        for e in memberdocumenters:
            section = self.member_sections.get(
                e[0].member_order, 'Miscellaneous')
            if self.env.app:
                e[0].parse_name()
                e[0].import_object()
                if members_check_module and not e[0].check_module():
                    continue
                user_section = self.env.app.emit_firstresult(
                    'autodocsumm-grouper', self.objtype, e[0].object_name,
                    e[0].object, section, self.object)
                section = user_section or section
            if not use_sections or section in use_sections:
                documenters.setdefault(section, []).append(e)
        self.options.update(options_save)
        return documenters

    def add_autosummary(self, relative_ref_paths=False):
        """Add the autosammary table of this documenter.

        Parameters
        ==========
        relative_ref_paths: bool
            Use paths relative to the current module instead of
            absolute import paths for each object
        """
        if (
            self.options.get("autosummary")
            and not self.options.get("no-autosummary")
        ):

            grouped_documenters = self.get_grouped_documenters()

            sourcename = self.get_sourcename()

            for section, documenters in grouped_documenters.items():
                if not self.options.autosummary_no_titles:
                    self.add_line('**%s:**' % section, sourcename)

                self.add_line('', sourcename)

                self.add_line('.. autosummary::', sourcename)
                if self.options.autosummary_nosignatures:
                    self.add_line('    :nosignatures:', sourcename)
                self.add_line('', sourcename)
                indent = '    '

                for (documenter, _) in documenters:
                    obj_ref_path = documenter.fullname
                    if relative_ref_paths:
                        modname = self.modname + "."
                        if documenter.fullname.startswith(modname):
                            obj_ref_path = documenter.fullname[len(modname):]

                    self.add_line(indent + '~' + obj_ref_path, sourcename)

                self.add_line('', sourcename)


class AutoSummModuleDocumenter(ModuleDocumenter, AutosummaryDocumenter):
    """Module documentor with autosummary tables of its members.

    This class has the same functionality as the base
    :class:`sphinx.ext.autodoc.ModuleDocumenter` class but with an additional
    `autosummary`.
    It's priority is slightly higher than the one of the ModuleDocumenter.
    """

    #: slightly higher priority than
    #: :class:`sphinx.ext.autodoc.ModuleDocumenter`
    priority = ModuleDocumenter.priority + 0.1

    #: original option_spec from :class:`sphinx.ext.autodoc.ModuleDocumenter`
    #: but with additional autosummary boolean option
    option_spec = ModuleDocumenter.option_spec.copy()
    option_spec['autosummary'] = bool_option
    option_spec['autosummary-no-nesting'] = bool_option
    option_spec['autosummary-sections'] = list_option
    option_spec['autosummary-no-titles'] = bool_option
    option_spec['autosummary-force-inline'] = bool_option
    option_spec['autosummary-nosignatures'] = bool_option

    #: Add options for members for the autosummary
    for _option in member_options.intersection(option_spec):
        option_spec['autosummary-' + _option] = option_spec[_option]
    del _option

    member_sections = {
        ad.ClassDocumenter.member_order: 'Classes',
        ad.ExceptionDocumenter.member_order: 'Exceptions',
        ad.FunctionDocumenter.member_order: 'Functions',
        ad.DataDocumenter.member_order: 'Data',
    }
    """:class:`dict` that includes the autosummary sections

    This dictionary defines the sections for the autosummmary option. The
    values correspond to the :attr:`sphinx.ext.autodoc.Documenter.member_order`
    attribute that shall be used for each section."""

    def add_content(self, *args, **kwargs):
        super().add_content(*args, **kwargs)

        self.add_autosummary(relative_ref_paths=True)

        if self.options.get("autosummary-no-nesting"):
            self.options["no-autosummary"] = "True"


class AutoSummClassDocumenter(ClassDocumenter, AutosummaryDocumenter):
    """Class documentor with autosummary tables for its members.

    This class has the same functionality as the base
    :class:`sphinx.ext.autodoc.ClassDocumenter` class but with an additional
    `autosummary` option to provide the ability to provide a summary of all
    methods and attributes.
    It's priority is slightly higher than the one of the ClassDocumenter
    """

    #: slightly higher priority than
    #: :class:`sphinx.ext.autodoc.ClassDocumenter`
    priority = ClassDocumenter.priority + 0.1

    #: original option_spec from :class:`sphinx.ext.autodoc.ClassDocumenter`
    #: but with additional autosummary boolean option
    option_spec = ClassDocumenter.option_spec.copy()
    option_spec['autosummary'] = bool_option
    option_spec['autosummary-no-nesting'] = bool_option
    option_spec['autosummary-sections'] = list_option
    option_spec['autosummary-no-titles'] = bool_option
    option_spec['autosummary-force-inline'] = bool_option
    option_spec['autosummary-nosignatures'] = bool_option

    #: Add options for members for the autosummary
    for _option in member_options.intersection(option_spec):
        option_spec['autosummary-' + _option] = option_spec[_option]
    del _option

    member_sections = {
        ad.ClassDocumenter.member_order: 'Classes',
        ad.MethodDocumenter.member_order: 'Methods',
        ad.AttributeDocumenter.member_order: 'Attributes',
    }
    """:class:`dict` that includes the autosummary sections

    This dictionary defines the sections for the autosummmary option. The
    values correspond to the :attr:`sphinx.ext.autodoc.Documenter.member_order`
    attribute that shall be used for each section."""

    def add_content(self, *args, **kwargs):
        super().add_content(*args, **kwargs)

        self.add_autosummary(relative_ref_paths=True)


class CallableDataDocumenter(DataDocumenter):
    """:class:`sphinx.ext.autodoc.DataDocumenter` that uses the __call__ attr
    """

    priority = DataDocumenter.priority + 0.1

    def format_args(self):
        # for classes, the relevant signature is the __init__ method's
        callmeth = self.get_attr(self.object, '__call__', None)
        if callmeth is None:
            return None
        return process_signature(callmeth)

    def get_doc(self, *args, **kwargs):
        """Reimplemented  to include data from the call method"""
        content = self.env.config.autodata_content
        if content not in ('both', 'call') or not self.get_attr(
                self.get_attr(self.object, '__call__', None), '__doc__'):
            return super(CallableDataDocumenter, self).get_doc(
                *args, **kwargs
            )

        # for classes, what the "docstring" is can be controlled via a
        # config value; the default is both docstrings
        docstrings = []
        if content != 'call':
            docstring = self.get_attr(self.object, '__doc__', None)
            docstrings = [docstring + '\n'] if docstring else []
        calldocstring = self.get_attr(
            self.get_attr(self.object, '__call__', None), '__doc__')
        if docstrings:
            docstrings[0] += calldocstring
        else:
            docstrings.append(calldocstring + '\n')

        doc = []
        for docstring in docstrings:
            if not isinstance(docstring, str):
                docstring = force_decode(docstring, encoding)
            doc.append(prepare_docstring(docstring, ignore))

        return doc


class CallableAttributeDocumenter(AttributeDocumenter):
    """:class:`sphinx.ext.autodoc.AttributeDocumenter` that uses the __call__
    attr
    """

    priority = AttributeDocumenter.priority + 0.1

    def format_args(self):
        # for classes, the relevant signature is the __init__ method's
        callmeth = self.get_attr(self.object, '__call__', None)
        if callmeth is None:
            return None
        return process_signature(callmeth)

    def get_doc(self, *args, **kwargs):
        """Reimplemented  to include data from the call method"""
        content = self.env.config.autodata_content
        if content not in ('both', 'call') or not self.get_attr(
                self.get_attr(self.object, '__call__', None), '__doc__'):
            return super(CallableAttributeDocumenter, self).get_doc(
                *args, **kwargs
            )

        # for classes, what the "docstring" is can be controlled via a
        # config value; the default is both docstrings
        docstrings = []
        if content != 'call':
            docstring = self.get_attr(self.object, '__doc__', None)
            docstrings = [docstring + '\n'] if docstring else []
        calldocstring = self.get_attr(
            self.get_attr(self.object, '__call__', None), '__doc__')
        if docstrings:
            docstrings[0] += calldocstring
        else:
            docstrings.append(calldocstring + '\n')

        doc = []
        if sphinx_version < [4, 0]:
            encoding = _get_arg("encoding", 0, None, *args, **kwargs)
            ignore = _get_arg("ignore", 1, 1, *args, **kwargs)
            for docstring in docstrings:
                if not isinstance(docstring, str):
                    docstring = force_decode(docstring, encoding)
                doc.append(prepare_docstring(docstring, ignore))
        else:
            for docstring in docstrings:
                doc.append(prepare_docstring(docstring))

        return doc


def dont_document_data(config, fullname):
    """Check whether the given object should be documented

    Parameters
    ----------
    config: sphinx.Options
        The configuration
    fullname: str
        The name of the object

    Returns
    -------
    bool
        Whether the data of `fullname` should be excluded or not"""
    if config.document_data is True:
        document_data = [re.compile('.*')]
    else:
        document_data = config.document_data
    if config.not_document_data is True:
        not_document_data = [re.compile('.*')]
    else:
        not_document_data = config.not_document_data
    return (
            # data should not be documented
            (any(re.match(p, fullname) for p in not_document_data)) or
            # or data is not included in what should be documented
            (not any(re.match(p, fullname) for p in document_data)))


class NoDataDataDocumenter(CallableDataDocumenter):
    """DataDocumenter that prevents the displaying of large data"""

    #: slightly higher priority as the one of the CallableDataDocumenter
    priority = CallableDataDocumenter.priority + 0.1

    def __init__(self, *args, **kwargs):
        super(NoDataDataDocumenter, self).__init__(*args, **kwargs)
        fullname = '.'.join(self.name.rsplit('::', 1))
        if hasattr(self.env, 'config') and dont_document_data(
                self.env.config, fullname):
            self.options = Options(self.options)
            self.options.annotation = ' '


class NoDataAttributeDocumenter(CallableAttributeDocumenter):
    """AttributeDocumenter that prevents the displaying of large data"""

    #: slightly higher priority as the one of the CallableAttributeDocumenter
    priority = CallableAttributeDocumenter.priority + 0.1

    def __init__(self, *args, **kwargs):
        super(NoDataAttributeDocumenter, self).__init__(*args, **kwargs)
        fullname = '.'.join(self.name.rsplit('::', 1))
        if hasattr(self.env, 'config') and dont_document_data(
                self.env.config, fullname):
            self.options = Options(self.options)
            self.options.annotation = ' '


class AutoDocSummDirective(SphinxDirective):
    """A directive to generate an autosummary.

    Usage::

        .. autoclasssum:: <Class>

        .. automodsum:: <module>

    The directive additionally supports all options of the ``autoclass`` or
    ``automod`` directive respectively. Sections can be a list of section titles
    to be included. If ommitted, all sections are used.
    """

    has_content = False

    option_spec = AutodocDirective.option_spec

    required_arguments = 1
    optional_arguments = 0

    def run(self):
        reporter = self.state.document.reporter

        try:
            source, lineno = reporter.get_source_and_line(self.lineno)
        except AttributeError:
            source, lineno = (None, None)

        # look up target Documenter
        objtype = self.name[4:-4]  # strip prefix (auto-) and suffix (-summ).
        doccls = self.env.app.registry.documenters[objtype]

        self.options['autosummary-force-inline'] = "True"
        self.options['autosummary'] = "True"
        if 'no-members' not in self.options:
            self.options['members'] = ""

        # process the options with the selected documenter's option_spec
        try:
            documenter_options = process_documenter_options(doccls, self.config,
                                                            self.options)
        except (KeyError, ValueError, TypeError) as exc:
            # an option is either unknown or has a wrong type
            logger.error(
                'An option to %s is either unknown or has an invalid '
                'value: %s', self.name, exc,
                location=(self.env.docname, lineno))
            return []

        # generate the output
        params = DocumenterBridge(self.env, reporter, documenter_options,
                                  lineno, self.state)
        documenter = doccls(params, self.arguments[0])
        documenter.add_autosummary()

        node = nodes.paragraph()
        node.document = self.state.document
        self.state.nested_parse(params.result, 0, node)

        return node.children


def setup(app):
    """setup function for using this module as a sphinx extension"""
    app.setup_extension('sphinx.ext.autosummary')
    app.setup_extension('sphinx.ext.autodoc')
    app.add_directive('autoclasssumm', AutoDocSummDirective)
    app.add_directive('automodulesumm', AutoDocSummDirective)

    AUTODOC_DEFAULT_OPTIONS.extend(
        [option for option in AutoSummModuleDocumenter.option_spec
         if option not in AUTODOC_DEFAULT_OPTIONS])

    AUTODOC_DEFAULT_OPTIONS.extend(
        [option for option in AutoSummClassDocumenter.option_spec
         if option not in AUTODOC_DEFAULT_OPTIONS])

    # make sure to allow inheritance when registering new documenters
    registry = app.registry.documenters
    for cls in [AutoSummClassDocumenter, AutoSummModuleDocumenter,
                CallableAttributeDocumenter, NoDataDataDocumenter,
                NoDataAttributeDocumenter]:
        if not issubclass(registry.get(cls.objtype), cls):
            app.add_autodocumenter(cls, override=True)

    # group event
    app.add_event('autodocsumm-grouper')

    # config value
    app.add_config_value('autodata_content', 'class', True)
    app.add_config_value('document_data', True, True)
    app.add_config_value('not_document_data', [], True)
    return {'version': sphinx.__display_version__, 'parallel_read_safe': True}





"""Test module for autodocsumm.

**Disclaimer**

Copyright 2016-2019, Philipp S. Sommer
Copyright 2020-2021, Helmholtz-Zentrum Hereon

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
import bs4
import pytest
import sphinx


from autodocsumm import sphinx_version


def in_between(full, sub, s0, *others):
    i0 = full.index(s0)
    try:
        last = min(filter(lambda i: i > i0, map(full.index, others)))
    except ValueError:  # empty sequence
        last = len(full)
    return full.index(sub) > i0 and full.index(sub) < last


def get_html(app, fname):
    with open(app.outdir + '/' + fname) as f:
        return f.read()


def in_autosummary(what, html) -> bool:
    soup = bs4.BeautifulSoup(html, "html.parser")
    autosummaries = soup("table")
    found = False
    for tag in autosummaries:
        if tag.find_all("span", string=what):
            found = True
            break
    return found


class TestAutosummaryDocumenter:

    def test_module(self, app):
        app.build()
        html = get_html(app, 'test_module.html')
        assert in_autosummary("TestClass", html)
        assert in_autosummary("test_func", html)
        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        # test whether the right objects are included
        assert in_autosummary("class_caller", html)
        assert 'Caller docstring for class attribute' in html

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        try:
            assert 'Should be included' in html
        except AssertionError: # sphinx>=3.5
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">be</span> '
                '<span class="pre">included\'</span>'
            ) in html
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">be</span> '
                '<span class="pre">skipped\'</span>'
            ) not in html
        else:
            assert 'Should be skipped' not in html

        try:
            assert 'Should also be included' in html
        except AssertionError: # sphinx>=3.5
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">also</span> '
                '<span class="pre">be</span> '
                '<span class="pre">included\'</span>'
            ) in html
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">also</span> '
                '<span class="pre">be</span> '
                '<span class="pre">skipped\'</span>'
            ) not in html
        else:
            assert 'Should also be skipped' not in html

    def test_module_no_nesting(self, app):
        app.build()
        html = get_html(app, 'test_module_no_nesting.html')

        assert in_autosummary("TestClass", html)
        assert in_autosummary("test_func", html)

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        # test that elements of TestClass are not autosummarized,
        # since nesting is disabled.
        assert not in_autosummary("test_method", html)
        assert not in_autosummary("test_attr", html)

        # test the members are still displayed
        assert re.search(
            r'<dt( class=".*")? id="dummy.Class_CallTest"( class=".*")*>',
            html,
        )

    def test_module_summary_only(self, app):
        app.build()
        html = get_html(app, 'test_module_summary_only.html')
        assert in_autosummary("TestClass", html)
        assert in_autosummary("test_func", html)

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        assert not re.search(
            r'<dt( class=".*")? id="dummy.Class_CallTest"( class=".*")*>',
            html,
        )

    def test_module_with_title(self, app):
        app.build()
        html = get_html(app, 'test_module_title.html')
        assert in_autosummary("TestClass", html)
        assert in_autosummary("test_func", html)
        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        # test whether the right objects are included
        assert in_autosummary("class_caller", html)
        assert 'Caller docstring for class attribute' in html

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)
        try:
            assert 'Should be included' in  html
        except AssertionError: # sphinx>=3.5
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">be</span> '
                '<span class="pre">included\'</span>'
            ) in html
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">be</span> '
                '<span class="pre">skipped\'</span>'
            ) not in html
        else:
            assert 'Should be skipped' not in html
        try:
            assert 'Should also be included' in html
        except AssertionError: # sphinx>=3.5
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">also</span> '
                '<span class="pre">be</span> '
                '<span class="pre">included\'</span>'
            ) in html
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">also</span> '
                '<span class="pre">be</span> '
                '<span class="pre">skipped\'</span>'
            ) not in html
        else:
            assert 'Should also be skipped' not in html

    def test_module_nosignatures(self, app):
        app.build()

        html = get_html(app, 'test_module_nosignatures.html')
        assert in_autosummary("TestClass", html)
        assert in_autosummary("test_func", html)

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        assert not re.search(
            r'<dt( class=".*")? id="dummy.Class_CallTest"( class=".*")*>',
            html,
        )
        assert '()' not in html

    def test_class(self, app):
        app.build()
        html = get_html(app, '/test_class.html')

        if sphinx_version[:2] > [3, 1]:
            assert in_autosummary("instance_attribute", html)
        elif sphinx_version[:2] < [3, 1]:
            assert in_autosummary("TestClass.instance_attribute", html)

        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        # test escaping of *
        assert r'\*args' not in html
        assert r', \*\*kwargs' not in html
        assert '*args' in html
        assert '**kwargs' in html

        # test whether the right objects are included
        assert in_autosummary("class_caller", html)
        assert 'Caller docstring for class attribute' in html

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        assert 'Should be skipped' not in html
        try:
            assert 'Should be included' in html
        except AssertionError: # sphinx>=3.5
            assert (
                '<span class="pre">\'Should</span> '
                '<span class="pre">be</span> '
                '<span class="pre">included\'</span>'
            ) in html

        assert 'DummySection' in html
        assert in_between(
            html, '<span class="pre">class_caller</span>', 'DummySection',
            'Attributes', 'Methods'
        ), 'class_caller Attribute not in the right Section!'

        # check if the InnerClass is in the Classes section (which ends with
        # the DummySection)
        assert in_between(
            html, '<span class="pre">InnerClass</span>', 'Classes',
            'DummySection'
        )

    @pytest.mark.skipif(
        sphinx_version[:2] < [3, 1], reason="Only available for sphinx>=3"
    )
    def test_class_order(self, app):
        app.build()
        html = get_html(app, '/test_class_order.html')

        if sphinx_version[:2] > [3, 1]:
            assert in_autosummary("instance_attribute", html)
        elif sphinx_version[:2] < [3, 1]:
            assert in_autosummary("TestClass.instance_attribute", html)

        assert in_autosummary("test_attr", html)
        assert in_autosummary("large_data", html)

        assert (
            html.index('<span class="pre">test_attr</span>')
            < html.index('<span class="pre">large_data</span>')
        )

    def test_class_summary_only(self, app):
        app.build()
        html = get_html(app, '/test_class_summary_only.html')

        if sphinx_version[:2] > [3, 1]:
            assert in_autosummary("instance_attribute", html)
        elif sphinx_version[:2] < [3, 1]:
            assert in_autosummary("TestClass.instance_attribute", html)

        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        # test whether the right objects are included
        assert in_autosummary("class_caller", html)

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        assert not re.search(
            r'<dt( class=".*")? id="dummy.TestClass.small_data"( class=".*")*>',
            html,
        )

    def test_class_nosignatures(self, app):
        app.build()
        html = get_html(app, '/test_class_nosignatures.html')

        if sphinx_version[:2] > [3, 1]:
            assert in_autosummary("instance_attribute", html)
        elif sphinx_version[:2] < [3, 1]:
            assert in_autosummary("TestClass.instance_attribute", html)

        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        # test whether the right objects are included
        assert in_autosummary("class_caller", html)

        # test whether the data is shown correctly
        assert in_autosummary("large_data", html)
        assert in_autosummary("small_data", html)

        assert not re.search(
            r'<dt( class=".*")? id="dummy.TestClass.small_data"( class=".*")*>',
            html,
        )

        assert '()' not in html

    def test_inherited(self, app):
        app.build()
        html = get_html(app, '/test_inherited.html')
        assert in_autosummary("test_method", html)

    @pytest.mark.xfail
    def test_warnings_depreciation(self, app):
        with pytest.warns(
            sphinx.deprecation.RemovedInSphinx40Warning,
            r'(?s).*Autosummary.warnings',
        ):
            app.build()

    def test_autoclasssumm_inline(self, app):
        """Test an AutoDocSummDirective inline."""
        app.build()

        html = get_html(app, '/test_autoclasssumm_inline.html')

        methods_title = "<strong>Methods:</strong>"

        num_section_findings = len(re.findall(methods_title, html))

        assert num_section_findings == 1

        methods_start = html.index(methods_title)
        docstring_end = html.index("This is after the summary")

        assert docstring_end > methods_start

    def test_class_submodule(self, app):
        app.build()

        html = get_html(app, '/test_class_submodule.html')

        # check that hyperlink for instance method exists in summary table
        assert re.findall(r'<td>.*href="#dummy_submodule\.submodule1'
                          r'\.SubmoduleClass1\.func1".*</td>', html)

    def test_module_submodule(self, app):
        app.build()

        html = get_html(app, '/test_module_submodule.html')

        # check that hyperlink for class exists in summary table
        assert re.findall(r'<td>.*href="#dummy_submodule\.submodule2'
                          r'\.SubmoduleClass2".*</td>', html)

        # check that hyperlink for instance method exists in summary table
        assert re.findall(r'<td>.*href="#dummy_submodule\.submodule2'
                          r'\.SubmoduleClass2\.func2".*</td>', html)


class TestAutoDocSummDirective:
    """Test case for the :class:`autodocsumm.AutoDocSummDirective`."""

    def test_autoclasssumm(self, app):
        """Test building the autosummary of a class."""
        app.build()

        html = get_html(app, '/test_autoclasssumm.html')

        # the class docstring must not be in the html
        assert "Class test for autosummary" not in html

        # test if the methods and attributes are there in a table
        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

    def test_autoclasssumm_no_titles(self, app):
        """Test building the autosummary of a class."""
        app.build()

        html = get_html(app, '/test_autoclasssumm_no_titles.html')

        # the class docstring must not be in the html
        assert "Class test for autosummary" not in html

        # test if the methods and attributes are there in a table
        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        assert "<strong>Methods</strong>" not in html

    def test_autoclasssumm_some_sections(self, app):
        """Test building the autosummary of a class with some sections only."""
        app.build()

        html = get_html(app, '/test_autoclasssumm_some_sections.html')

        # the class docstring must not be in the html
        assert "Class test for autosummary" not in html

        # test if the methods and attributes are there in a table
        assert not in_autosummary("test_method", html)
        assert in_autosummary("class_caller", html)
        assert in_autosummary("test_attr", html)

    def test_autoclasssumm_nosignatures(self, app):
        """Test building the autosummary of a class without signatures."""
        app.build()

        html = get_html(app, '/test_autoclasssumm_nosignatures.html')

        # the class docstring must not be in the html
        assert "Class test for autosummary" not in html

        # test if the methods and attributes are there in a table
        assert in_autosummary("test_method", html)
        assert in_autosummary("test_attr", html)

        assert '()' not in html

    def test_automodulesumm(self, app):
        """Test building the autosummary of a module."""
        app.build()

        html = get_html(app, '/test_automodulesumm.html')

        # the class docstring must not be in the html
        assert "Module for testing the autodocsumm" not in html

        # test if the classes, data and functions are there in a table
        assert in_autosummary("Class_CallTest", html)
        assert in_autosummary("large_data", html)
        assert in_autosummary("test_func", html)

    def test_automodulesumm_some_sections(self, app):
        """Test building the autosummary of a module with some sections only."""
        app.build()

        html = get_html(app, '/test_automodulesumm_some_sections.html')

        # the class docstring must not be in the html
        assert "Module for testing the autodocsumm" not in html

        # test if the classes, data and functions are there in a table
        assert not in_autosummary("Class_CallTest", html)
        assert in_autosummary("large_data", html)
        assert in_autosummary("test_func", html)

    def test_automodulesumm_nosignatures(self, app):
        """Test building the autosummary of a module without signatures."""
        app.build()

        html = get_html(app, '/test_automodulesumm_nosignatures.html')

        # the class docstring must not be in the html
        assert "Module for testing the autodocsumm" not in html

        # test if the classes, data and functions are there in a table
        assert in_autosummary("Class_CallTest", html)
        assert in_autosummary("large_data", html)
        assert in_autosummary("test_func", html)

        assert '()' not in html

    def test_empty(self, app):
        app.build()

        html = get_html(app, '/test_empty.html')

        assert not in_autosummary("product", html)



# -*- coding: utf-8 -*-
"""
Module for testing the autodocsumm
----------------------------------

Just a dummy module with some class definitions
"""


#: to test if the data is included
test_data = None


def test_func():
    """Test if function is contained in autosummary"""
    pass


class Class_CallTest(object):
    """A class defining a __call__ method"""

    def __get__(self, instance, owner):
        return self

    def __set__(self, instance, value):
        """Actually not required. We just implement it to ensure the python
        "help" function works well"""
        pass

    def __call__(self, a, b):
        """
        Caller docstring for class attribute

        Parameters
        ----------
        a: any
            dummy parameter
        b: anything else
            second dummy parameter"""
        pass


class TestClass(object):
    """Class test for autosummary"""

    def test_method(self):
        """Test if the method is included"""
        pass

    #: to test if the class attribute is included
    test_attr = None

    class_caller = Class_CallTest()

    #: data to be included
    large_data = 'Should be included'

    #: data to be skipped
    small_data = 'Should be skipped'


#: data to be skipped
large_data = 'Should also be skipped'

#: data to be included
small_data = 'Should also be included'



# -*- coding: utf-8 -*-
import sys
import sphinx
sys.path.insert(0, '.')

master_doc = 'index'

extensions = ['autodocsumm']

if tuple(map(int, sphinx.__version__.split('.')[:2])) < (2, 1):
    autodoc_default_flags = ['show_inheritance', 'autosummary', 'members']
else:
    autodoc_default_options = {
        'show_inheritance': True,
        'autosummary': True,
        'members': True
        }

not_document_data = [
    'dummy.large_data', 'dummy.TestClass.small_data',
    'dummy_title.large_data', 'dummy_title.TestClass.small_data']

autodoc_typehints = "description"

autodata_content = 'both'


def member_filter(app, what, name, obj, skip, options):
    import dummy
    import dummy_title
    if (obj is dummy.TestClass.class_caller or
            obj is dummy_title.TestClass.class_caller):
        return False
    return None


def group_member(app, what, name, obj, section, parent):
    if name == 'class_caller':
        return 'DummySection'


def setup(app):
    app.connect('autodoc-skip-member', member_filter)
    app.connect('autodocsumm-grouper', group_member)
    app.setup_extension('sphinx.ext.napoleon')



import os.path as osp
import sys
import pytest

from sphinx.testing.path import path

pytest_plugins = 'sphinx.testing.fixtures'



sphinx_supp = osp.abspath(osp.join(osp.dirname(__file__), "tests"))


@pytest.fixture(scope='session')
def rootdir():
    return path(sphinx_supp)


sys.path.insert(0, osp.join(sphinx_supp, "test-root"))



class MyClass(object):
    """Some class

    .. autoclasssumm:: MyClass
        :autosummary-sections: Methods

    This will be after the autosummary table.
    """

    def do_something(self):
        """Do something"""
        pass

    #: Any instance attribute
    some_attr = None



"""
Some module
-----------

Just a dummy module with some class definition
"""


class CallableDescriptor(object):
    """A class defining a __call__ method"""

    def __get__(self, instance, owner):
        return self

    def __set__(self, instance, value):
        """Actually not required. We just implement it to ensure the python
        "help" function works well"""
        pass

    def __call__(self, a, b):
        """
        Caller docstring for class attribute

        Parameters
        ----------
        a: any
            dummy parameter
        b: anything else
            second dummy parameter"""
        pass


class MyClass(object):
    """Some class

    With some description"""

    do_something = CallableDescriptor()



# -*- coding: utf-8 -*-
"""
Module for testing the autodocsumm

Just a dummy module with some class definitions
"""


#: to test if the data is included
test_data = None


def test_func():
    """Test if function is contained in autosummary"""
    pass


class Class_CallTest(object):
    """A class defining a __call__ method"""

    def __get__(self, instance, owner):
        return self

    def __set__(self, instance, value):
        """Actually not required. We just implement it to ensure the python
        "help" function works well"""
        pass

    def __call__(self, a, b):
        """
        Caller docstring for class attribute

        Parameters
        ----------
        a: any
            dummy parameter
        b: anything else
            second dummy parameter"""
        pass


class TestClass(object):
    """Class test for autosummary"""

    def __init__(self):
        #: This is an instance attribute
        self.instance_attribute = 1

    def test_method(self):
        """Test if the method is included"""
        pass

    def test_method_args_kwargs(self, *args, **kwargs):
        """The stars at args and kwargs should not be escaped"""

    class InnerClass(object):
        """A test for an inner class"""

    #: to test if the class attribute is included
    test_attr = None

    class_caller = Class_CallTest()

    #: data to be included
    large_data = 'Should be included'

    #: data to be skipped
    small_data = 'Should be skipped'


class InheritedTestClass(TestClass):
    """Class test for inherited attributes"""


class TestClassWithInlineAutoClassSumm:
    """Class test for the autodocsummary inline

    .. autoclasssumm:: TestClassWithInlineAutoClassSumm

    This is after the summary.
    """

    def test_method_of_inline_test(self):
        """A test method."""
        pass


#: data to be skipped
large_data = 'Should also be skipped'

#: data to be included
small_data = 'Should also be included'



"""
Some module
-----------

Just a dummy module with some class definition
"""


class MyClass(object):
    """Some class

    With some description"""

    def do_something(self):
        """Do something"""
        pass

    #: Any instance attribute
    some_attr = None

    #: Any other instance attribute
    some_other_attr = None


#: Some module data
large_data = 'Whatever'
